/**
 * IcecastPlayer - Audio player with visualization for Icecast streams
 */
class IcecastPlayer {
    // Constants
    static SMOOTHING_FACTOR = 0.5;
    static ACTIVITY_THRESHOLD = 0.5;
    static RELEASE_THRESHOLD = 1.5; // Lower threshold for keeping frequencies active (hysteresis)
    static FREQ_PERSISTENCE = 8; // Increased from 3 to 8 - much more consecutive frames required
    static HIGH_FREQ_DAMPENING = 0.7; // Apply additional dampening factor to high frequencies
    static MIN_RANGE_WIDTH = 30;
    static RANGE_ADAPTATION_SPEED = 0.05;
    static RANGE_EXPANSION_SPEED = 0.01;
    static VERTICAL_SCALE = 0.9;
    static RECOVERY_DEBOUNCE = 1000;
    static MAX_RECONNECT_ATTEMPTS = 1000;
    static INITIAL_RECONNECT_DELAY = 1000; // 1 second
    static AVG_BUFFER_SIZE = 10; // Number of frames to average for the stroke
    
    /**
     * Constructor - initialize the player
     * @param {string} audioElementId - The ID of the audio element
     * @param {string} canvasId - The ID of the canvas element for visualization
     */
    constructor({ audioElementId, canvasId }) {
        // Elements
        this.audioElement = document.getElementById(audioElementId);
        this.canvas = document.getElementById(canvasId);
        
        // Only proceed if we have both required elements
        if (!this.audioElement || !this.canvas) {
            console.warn('Required elements not found. Audio player initialization skipped.');
            return;
        }
        
        // Create a 2D context with anti-aliasing enabled
        this.ctx = this.canvas.getContext('2d', { alpha: true, antialias: true });
        if (!this.ctx) {
            console.warn('Could not get canvas context. Visualization disabled.');
            return;
        }
        
        // Enable image smoothing
        this.ctx.imageSmoothingEnabled = true;
        this.ctx.imageSmoothingQuality = 'high';
        
        // Audio context and analyzer
        this.audioContext = null;
        this.analyser = null;
        this.dataArray = null;
        this.source = null;
        
        // Animation and state
        this.animationId = null;
        this.isPlaying = false;
        this.isBuffering = true; // Start in buffering state
        this.attemptedAutoplay = false;
        this.hasBeenManuallyPaused = false;
        
        // Previous values for smoothing
        this.previousData = [];
        
        // Short-term average buffer for engineering-style visualization
        this.avgBuffer = [];
        this.avgBufferIndex = 0;
        
        // Frequency range tracking
        this.activeFreqStart = 0;
        this.activeFreqEnd = 0;
        this.spectrumFitted = false;
        
        // Frequency persistence tracking for noise gate
        this.freqPersistence = [];
        
        // Recovery variables
        this.lastRecoveryAttempt = 0;
        this.recoveryInProgress = false;
        this.reconnectAttempts = 0;
        this.reconnectDelay = IcecastPlayer.INITIAL_RECONNECT_DELAY;
        this.reconnectTimeout = null;
        this.lastPlaybackTime = 0;
        
        // Initialize
        this.initializeAudioContext();
        this.setupEventListeners();
        this.resizeCanvas();
        this.updatePlaybackUI();
        
        // Try to auto-play (will likely require user interaction in most browsers)
        setTimeout(() => this.playAudio(), 0);
    }
    
    /**
     * Initialize audio context in suspended state
     */
    initializeAudioContext() {
        try {
            this.audioContext = new (window.AudioContext || window.webkitAudioContext)();
            console.log('AudioContext created with initial state:', this.audioContext.state);
        } catch (e) {
            console.error('Web Audio API is not supported in this browser', e);
        }
    }
    
    /**
     * Set up all event listeners
     */
    setupEventListeners() {
        // Handle canvas interaction - allow clicking the waveform to toggle play/pause
        this.canvas.addEventListener('click', () => {
            if (this.isPlaying) {
                this.pauseAudio();
            } else {
                this.playAudio();
            }
        });
        
        // Handle window resize
        window.addEventListener('resize', () => this.resizeCanvas());
        
        // Improve user interaction detection for AudioContext
        document.addEventListener('click', (event) => {
            console.log('User interaction detected');
            
            // Skip autoplay if the click was on the canvas (handled by canvas click handler)
            if (event.target === this.canvas) {
                return;
            }
            
            // This will enable audio after user interaction
            if (this.audioContext && this.audioContext.state === 'suspended') {
                console.log('Resuming AudioContext from user interaction');
                this.audioContext.resume().then(() => {
                    console.log('AudioContext resumed successfully from user interaction');
                }).catch(e => {
                    console.error('Failed to resume AudioContext from user interaction', e);
                });
            }
            
            // Only try to play if we haven't manually paused
            if (!this.isPlaying && !this.hasBeenManuallyPaused) {
                this.playAudio();
            }
        }, { once: false });
        
        // Also listen for touchstart events (mobile devices)
        document.addEventListener('touchstart', () => {
            if (this.audioContext && this.audioContext.state === 'suspended') {
                console.log('Resuming AudioContext from touch interaction');
                this.audioContext.resume().catch(e => {
                    console.error('Failed to resume AudioContext from touch', e);
                });
            }
        }, { once: true });
        
        // Audio event listeners for buffering detection
        this.audioElement.addEventListener('waiting', () => this.startBuffering());
        this.audioElement.addEventListener('stalled', () => this.startBuffering());
        this.audioElement.addEventListener('suspended', () => this.startBuffering());
        this.audioElement.addEventListener('loadstart', () => this.startBuffering());
        this.audioElement.addEventListener('loadeddata', () => {
            if (this.isBuffering && this.audioElement.readyState >= 3) {
                this.stopBuffering();
            }
        });
        this.audioElement.addEventListener('playing', () => this.stopBuffering());
        this.audioElement.addEventListener('canplay', () => this.stopBuffering());
        this.audioElement.addEventListener('canplaythrough', () => this.stopBuffering());
        
        // Listen for custom events from stream-monitor.js
        this.audioElement.addEventListener('stream-monitor:playback-started', () => {
            console.log('Received playback started event from stream-monitor');
            this.isPlaying = true;
            this.hasBeenManuallyPaused = false;
            this.stopBuffering();
            
            // Update UI to reflect playing state
            this.updatePlaybackUI();
        });
        
        // Listen for stream problems detected by stream-monitor.js
        this.audioElement.addEventListener('stream-monitor:problem-detected', (event) => {
            console.log('Stream problem detected:', event.detail?.reason);
            
            // Only attempt recovery if we're actually playing and the stream needs recovery
            if (this.isPlaying && this.needsRecovery()) {
                this.attemptStreamRecovery();
            }
        });
        
        // Update event listeners to ensure they call attemptStreamRecovery
        this.audioElement.addEventListener('ended', () => {
            if (this.isPlaying && this.needsRecovery()) {
                this.isBuffering = false; // Reset buffering state
                this.attemptStreamRecovery();
            }
        });
        
        this.audioElement.addEventListener('error', () => {
            if (this.isPlaying && this.needsRecovery()) {
                this.isBuffering = false; // Reset buffering state
                this.attemptStreamRecovery();
            }
        });
        
        // Track playback progress to detect stalls
        setInterval(() => {
            if (!this.audioElement.paused && this.isPlaying) {
                // If currentTime hasn't changed in the last interval, we might be stalled
                if (this.audioElement.currentTime === this.lastPlaybackTime) {
                    console.log('Playback stalled, currentTime not advancing');
                    if (this.needsRecovery()) {
                        this.attemptStreamRecovery();
                    }
                }
                // Update the last time we saw playback progress
                this.lastPlaybackTime = this.audioElement.currentTime;
            }
        }, 5000);
    }
    
    /**
     * Fix canvas size
     */
    resizeCanvas() {
        const devicePixelRatio = window.devicePixelRatio || 1;
        const rect = this.canvas.getBoundingClientRect();
        
        // Set the canvas dimensions in actual pixels
        this.canvas.width = rect.width * devicePixelRatio;
        this.canvas.height = rect.height * devicePixelRatio;
        
        // Normalize the coordinate system
        this.ctx.scale(devicePixelRatio, devicePixelRatio);
        
        // Re-enable image smoothing after canvas resize (it resets on width/height changes)
        this.ctx.imageSmoothingEnabled = true;
        this.ctx.imageSmoothingQuality = 'high';
        
        // Store logical dimensions for drawing functions
        this.canvas.logicalWidth = rect.width;
        this.canvas.logicalHeight = rect.height;
        
        // Redraw if we're not playing (to ensure play button is visible)
        if (!this.isPlaying && !this.isBuffering) {
            this.updatePlaybackUI();
        }
    }
    
    /**
     * Helper function to detect Safari/iOS
     */
    isSafari() {
        return /^((?!chrome|android).)*safari/i.test(navigator.userAgent) || 
               /iPad|iPhone|iPod/.test(navigator.userAgent) || 
               (/Apple/.test(navigator.vendor) && !window.MSStream);
    }
    
    /**
     * Initialize audio context
     */
    setupAudioContext() {
        if (this.audioContext) {
            try {
                this.analyser = this.audioContext.createAnalyser();
                this.source = this.audioContext.createMediaElementSource(this.audioElement);
                
                this.analyser.fftSize = 1024; // Good balance between performance and detail
                const bufferLength = this.analyser.frequencyBinCount;
                this.dataArray = new Uint8Array(bufferLength);
                this.previousData = new Array(bufferLength).fill(0);
                
                // Initialize the average buffer for engineering-style visualization
                this.avgBuffer = new Array(IcecastPlayer.AVG_BUFFER_SIZE);
                for (let i = 0; i < IcecastPlayer.AVG_BUFFER_SIZE; i++) {
                    this.avgBuffer[i] = new Array(bufferLength).fill(0);
                }
                
                // Initialize the active frequency range
                this.activeFreqStart = 0;
                this.activeFreqEnd = bufferLength - 1;
                
                // Initialize frequency persistence tracking array
                this.freqPersistence = new Array(bufferLength).fill(0);
                
                this.source.connect(this.analyser);
                this.analyser.connect(this.audioContext.destination);
                
                return true;
            } catch (e) {
                console.error('Failed to setup audio context', e);
                return false;
            }
        }
        return false;
    }
    
    /**
     * Find the active frequency range with noise-gate like behavior
     */
    findActiveFrequencyRange(data) {
        if (!this.spectrumFitted && data.some(value => value > IcecastPlayer.ACTIVITY_THRESHOLD)) {
            // We have some audio data, initialize fitting
            this.spectrumFitted = true;
        }
        
        if (!this.spectrumFitted) {
            // Not enough audio data yet
            return { start: 0, end: data.length - 1 };
        }
        
        // Apply additional dampening to high frequencies
        const dampedData = [...data];
        const midPoint = Math.floor(data.length / 2);
        for (let i = midPoint; i < data.length; i++) {
            // Progressive dampening - more dampening as frequency gets higher
            const dampFactor = IcecastPlayer.HIGH_FREQ_DAMPENING * 
                (1 - ((data.length - i) / (data.length - midPoint)));
            dampedData[i] = dampedData[i] * (1 - dampFactor);
        }
        
        // Update frequency persistence tracking for noise-gate behavior
        for (let i = 0; i < data.length; i++) {
            // If frequency is active, increment its persistence counter
            if (dampedData[i] > IcecastPlayer.ACTIVITY_THRESHOLD) {
                this.freqPersistence[i] = Math.min(this.freqPersistence[i] + 1, IcecastPlayer.FREQ_PERSISTENCE + 5);
            } 
            // If frequency is below release threshold, decrease its persistence counter more rapidly
            else if (dampedData[i] < IcecastPlayer.RELEASE_THRESHOLD) {
                this.freqPersistence[i] = Math.max(this.freqPersistence[i] - 2, 0); // Faster decay
            }
            // Otherwise if between thresholds, maintain current persistence (hysteresis)
        }
        
        // Find the first and last indices where the frequency is persistently active
        let start = 0;
        let end = data.length - 1;
        
        // Find start (first frequency with persistent activity)
        for (let i = 0; i < data.length; i++) {
            if (this.freqPersistence[i] >= IcecastPlayer.FREQ_PERSISTENCE) {
                start = Math.max(0, i - 5); // Add a small buffer
                break;
            }
        }
        
        // Find end (last frequency with persistent activity)
        for (let i = data.length - 1; i >= 0; i--) {
            if (this.freqPersistence[i] >= IcecastPlayer.FREQ_PERSISTENCE) {
                end = Math.min(data.length - 1, i + 5); // Add a small buffer
                break;
            }
        }
        
        // Ensure minimum width
        if (end - start < IcecastPlayer.MIN_RANGE_WIDTH) {
            const center = Math.floor((start + end) / 2);
            start = Math.max(0, center - Math.floor(IcecastPlayer.MIN_RANGE_WIDTH / 2));
            end = Math.min(data.length - 1, center + Math.ceil(IcecastPlayer.MIN_RANGE_WIDTH / 2));
        }
        
        return { start, end };
    }
    
    /**
     * Create a centralized function to update the UI based on playback state
     */
    updatePlaybackUI() {
        if (this.animationId) {
            cancelAnimationFrame(this.animationId);
            this.animationId = null;
        }
        
        const width = this.canvas.logicalWidth || this.canvas.width;
        const height = this.canvas.logicalHeight || this.canvas.height;
        
        // Clear canvas
        this.ctx.clearRect(0, 0, width, height);
        
        if (this.isBuffering) {
            // If buffering, show loading animation
            this.drawLoadingAnimation();
            return;
        }
        
        if (this.isPlaying) {
            // If playing and we have analyzer, show waveform
            if (this.analyser && !this.isSafari()) {
                this.drawSmoothWaveform();
                return;
            }
            
            // Otherwise draw pause button
            const centerX = width * 0.5;
            const centerY = height * 0.5;
            const size = Math.min(width, height) * 0.2;
            
            // Button background
            this.ctx.beginPath();
            this.ctx.arc(centerX, centerY, size, 0, Math.PI * 2, false);
            this.ctx.fillStyle = 'rgba(255, 255, 255, 0.2)';
            this.ctx.fill();
            
            // Pause bars
            this.ctx.fillStyle = '#64f87b';
            const barWidth = size * 0.15;
            const barHeight = size * 0.5;
            const barSpacing = size * 0.2;
            
            this.ctx.fillRect(centerX - barSpacing/2 - barWidth, centerY - barHeight/2, barWidth, barHeight);
            this.ctx.fillRect(centerX + barSpacing/2, centerY - barHeight/2, barWidth, barHeight);
        } else {
            // Not playing, show play button
            this.drawPlayIndicator(width, height);
            
            // // If this was a failed autoplay attempt, show message
            // if (this.attemptedAutoplay && !this.hasBeenManuallyPaused) {
            //     this.ctx.font = '14px Arial';
            //     this.ctx.textAlign = 'center';
            //     this.ctx.fillStyle = '#f8ef64';
            //     this.ctx.fillText('Click to play audio', width / 2, height - 15);
            // }
        }
    }
    
    /**
     * Draw loading animation
     */
    drawLoadingAnimation() {
        if (this.animationId) {
            cancelAnimationFrame(this.animationId);
        }
        
        this.animationId = requestAnimationFrame(() => this.drawLoadingAnimation());
        
        const width = this.canvas.logicalWidth || this.canvas.width;
        const height = this.canvas.logicalHeight || this.canvas.height;
        
        this.ctx.clearRect(0, 0, width, height);
        
        // Draw pulsing loading animation
        const time = Date.now() * 0.001;
        const pulseSize = Math.sin(time * 2) * 0.2 + 0.8; // Value between 0.6 and 1.0
        
        // Calculate intensity based on time for loading animation
        const intensity = Math.min(1, (Math.sin(time * 3) + 1) / 2);
        
        // Create dynamic colors based on intensity - yellow to orange to red
        let topR, topG, topB;
        let bottomR, bottomG, bottomB;
        
        if (intensity < 0.7) {
            // Yellow to Orange (0-0.7 intensity) - extended yellow range
            const factor = intensity / 0.7; // Scale to 0-1 range within this segment
            topR = Math.round(248 + (factor * 7)); // Increase red slightly
            topG = Math.round(239 - (factor * 100)); // Decrease green
            topB = Math.round(100 - (factor * 100)); // Decrease blue
            
            bottomR = Math.round(232 + (factor * 18)); // Increase red slightly
            bottomG = Math.round(142 - (factor * 42)); // Decrease green
            bottomB = Math.round(60 - (factor * 60)); // Decrease blue
        } else {
            // Orange to Red (0.7-1.0 intensity) - compressed orange/red range
            const factor = (intensity - 0.7) / 0.3; // Scale to 0-1 range within this segment
            topR = Math.round(255); // Full red
            topG = Math.round(139 - (factor * 139)); // Decrease green to 0
            topB = Math.round(0); // No blue
            
            // More deeply red bottom colors for high intensity
            bottomR = Math.round(240 + (factor * 15)); // Maximum red value
            bottomG = Math.round(15 - (factor * 15)); // Almost no green
            bottomB = Math.round(15 - (factor * 15)); // Almost no blue for pure deep red
        }
        
        // Create gradient with dynamic colors
        const gradient = this.ctx.createLinearGradient(0, 0, 0, height);
        gradient.addColorStop(0, `rgb(${topR}, ${topG}, ${topB})`);  // Dynamic color at top
        gradient.addColorStop(1, `rgb(${bottomR}, ${bottomG}, ${bottomB})`);  // Dynamic color at bottom
        
        this.ctx.strokeStyle = gradient;
        this.ctx.lineWidth = 1; // Reduced from 2 to 1 for a thinner line
        this.ctx.beginPath();
        
        // Draw a loading wave
        this.ctx.moveTo(0, height / 2);
        
        // Create a loading animation with moving dots
        const dotCount = 5;
        const dotSpacing = width / (dotCount - 1);
        
        for (let i = 0; i < dotCount; i++) {
            const x = i * dotSpacing;
            const offset = (time * 2 + i * 0.3) % 2; // Varies between 0 and 2
            const y = height / 2 + Math.sin(offset * Math.PI) * 20 * pulseSize;
            
            if (i === 0) {
                this.ctx.moveTo(x, y);
            } else {
                const cpx = x - dotSpacing / 2;
                const cpy = height / 2;
                this.ctx.quadraticCurveTo(cpx, cpy, x, y);
            }
            
            // Draw dots at each peak with dynamic color
            this.ctx.fillStyle = 'rgba(248, 239, 100, 0.7)';  // Fixed yellowish color with transparency
            this.ctx.beginPath();
            this.ctx.arc(x, y, 4 * pulseSize, 0, Math.PI * 2);
            this.ctx.fill();
        }
    }
    
    /**
     * Draw play indicator more prominently
     */
    drawPlayIndicator(width, height) {
        // Use logical width/height if available or fall back to given parameters
        const w = this.canvas.logicalWidth || width;
        const h = this.canvas.logicalHeight || height;
        
        // Draw a semi-transparent background circle
        const centerX = w * 0.5; // Position exactly in center horizontally
        const centerY = h * 0.5; // Position exactly in center vertically
        const radius = Math.min(w, h) * 0.2; // Bigger size
        
        // Draw circle background
        this.ctx.beginPath();
        this.ctx.arc(centerX, centerY, radius, 0, Math.PI * 2, false);
        this.ctx.fillStyle = 'rgba(255, 255, 255, 0.2)';
        this.ctx.fill();
        
        // Draw play triangle
        const triangleSize = radius * 0.7;
        this.ctx.beginPath();
        this.ctx.fillStyle = '#64f87b';
        this.ctx.moveTo(centerX + triangleSize/2, centerY);
        this.ctx.lineTo(centerX - triangleSize/3, centerY - triangleSize/2);
        this.ctx.lineTo(centerX - triangleSize/3, centerY + triangleSize/2);
        this.ctx.closePath();
        this.ctx.fill();
    }
    
    /**
     * Draw smooth waveform
     */
    drawSmoothWaveform() {
        if (!this.analyser || this.isBuffering) return;
        
        if (this.animationId) {
            cancelAnimationFrame(this.animationId);
        }
        
        // Set up animation loop
        this.animationId = requestAnimationFrame(() => this.drawSmoothWaveform());
        
        // For Safari, only show play/pause buttons
        if (this.isSafari()) {
            this.updatePlaybackUI();
            return;
        }
        
        this.analyser.getByteFrequencyData(this.dataArray);
        
        const width = this.canvas.logicalWidth || this.canvas.width;
        const height = this.canvas.logicalHeight || this.canvas.height;
        
        this.ctx.clearRect(0, 0, width, height);
        
        // Apply smoothing
        for (let i = 0; i < this.dataArray.length; i++) {
            this.previousData[i] = this.previousData[i] * IcecastPlayer.SMOOTHING_FACTOR + this.dataArray[i] * (1 - IcecastPlayer.SMOOTHING_FACTOR);
        }
        
        // Store current frame in circular buffer for short-term average
        this.avgBuffer[this.avgBufferIndex] = [...this.previousData];
        this.avgBufferIndex = (this.avgBufferIndex + 1) % IcecastPlayer.AVG_BUFFER_SIZE;
        
        // Calculate the short-term average across all frames in buffer
        const avgData = new Array(this.previousData.length).fill(0);
        for (let i = 0; i < this.previousData.length; i++) {
            let sum = 0;
            for (let j = 0; j < IcecastPlayer.AVG_BUFFER_SIZE; j++) {
                sum += this.avgBuffer[j][i] || 0;
            }
            avgData[i] = sum / IcecastPlayer.AVG_BUFFER_SIZE;
        }
        
        // Find active frequency range with current data
        const currentRange = this.findActiveFrequencyRange(this.previousData);
        
        // Use asymmetric adaptation - slower for expansion, faster for contraction
        // This makes it quicker to narrow the display but slower to widen it
        const startAdaptSpeed = currentRange.start < this.activeFreqStart 
            ? IcecastPlayer.RANGE_EXPANSION_SPEED  // Expanding left boundary (slower)
            : IcecastPlayer.RANGE_ADAPTATION_SPEED;  // Contracting left boundary (faster)
            
        const endAdaptSpeed = currentRange.end > this.activeFreqEnd 
            ? IcecastPlayer.RANGE_EXPANSION_SPEED  // Expanding right boundary (slower)
            : IcecastPlayer.RANGE_ADAPTATION_SPEED;  // Contracting right boundary (faster)
        
        // Smooth the range adaptation with asymmetric speeds
        this.activeFreqStart = Math.round(this.activeFreqStart * (1 - startAdaptSpeed) + currentRange.start * startAdaptSpeed);
        this.activeFreqEnd = Math.round(this.activeFreqEnd * (1 - endAdaptSpeed) + currentRange.end * endAdaptSpeed);
        
        // Calculate average audio level for color intensity
        let totalLevel = 0;
        for (let i = this.activeFreqStart; i <= this.activeFreqEnd; i++) {
            totalLevel += this.previousData[i];
        }
        const avgLevel = totalLevel / (this.activeFreqEnd - this.activeFreqStart + 1);
        const intensity = Math.min(1, avgLevel / 255);
        
        // Create dynamic colors based on intensity - yellow to orange to red
        let topR, topG, topB;
        let bottomR, bottomG, bottomB;
        
        if (intensity < 0.7) {
            // Yellow to Orange (0-0.7 intensity) - extended yellow range
            const factor = intensity / 0.7; // Scale to 0-1 range within this segment
            topR = Math.round(248 + (factor * 7)); // Increase red slightly
            topG = Math.round(239 - (factor * 100)); // Decrease green
            topB = Math.round(100 - (factor * 100)); // Decrease blue
            
            bottomR = Math.round(232 + (factor * 18)); // Increase red slightly
            bottomG = Math.round(142 - (factor * 42)); // Decrease green
            bottomB = Math.round(60 - (factor * 60)); // Decrease blue
        } else {
            // Orange to Red (0.7-1.0 intensity) - compressed orange/red range
            const factor = (intensity - 0.7) / 0.3; // Scale to 0-1 range within this segment
            topR = Math.round(255); // Full red
            topG = Math.round(139 - (factor * 139)); // Decrease green to 0
            topB = Math.round(0); // No blue
            
            // More deeply red bottom colors for high intensity
            bottomR = Math.round(240 + (factor * 15)); // Maximum red value
            bottomG = Math.round(15 - (factor * 15)); // Almost no green
            bottomB = Math.round(15 - (factor * 15)); // Almost no blue for pure deep red
        }
        
        // Create gradient for the fill with dynamic colors
        const fillGradient = this.ctx.createLinearGradient(0, 0, 0, height);
        fillGradient.addColorStop(0, `rgba(248, 239, 100, 0.5)`);  // Fixed yellowish color with transparency
        fillGradient.addColorStop(1, `rgba(232, 142, 60, 0.1)`);  // Fixed orange color with more transparency
        
        // Create gradient for the stroke with dynamic colors
        const strokeGradient = this.ctx.createLinearGradient(0, 0, 0, height);
        strokeGradient.addColorStop(0, `rgb(${topR}, ${topG}, ${topB})`);  // Dynamic color at top
        strokeGradient.addColorStop(1, `rgb(${bottomR}, ${bottomG}, ${bottomB})`);  // Dynamic color at bottom
        
        // Draw the filled waveform using instantaneous data
        this.ctx.fillStyle = fillGradient;
        this.drawWaveformPath(this.previousData, width, height);
        this.ctx.fill();
        
        // Apply semi-transparency to the stroke
        this.ctx.globalAlpha = 0.6;
        
        // Draw the stroke using averaged data
        this.ctx.strokeStyle = strokeGradient;
        this.ctx.lineWidth = 1.25; // Slightly thicker line for better anti-aliasing
        this.ctx.lineCap = 'round';
        this.ctx.lineJoin = 'round';
        this.drawWaveformPath(avgData, width, height);
        this.ctx.stroke();
        
        // Reset globalAlpha
        this.ctx.globalAlpha = 1.0;
        
        // Draw play indicator if not playing
        if (!this.isPlaying) {
            this.drawPlayIndicator(width, height);
        }
    }
    
    /**
     * Helper method to draw waveform path with given data
     */
    drawWaveformPath(data, width, height) {
        // Create a smooth curve using only the active frequency range
        const frequencyRange = this.activeFreqEnd - this.activeFreqStart + 1;
        const sliceWidth = width / frequencyRange;
        let x = 0;
        
        // Start the path
        this.ctx.beginPath();
        
        // Align to half-pixels to reduce aliasing
        const alignToPixel = (value) => Math.round(value) + 0.5;
        
        // First point
        const firstPoint = (height - ((data[this.activeFreqStart] / 255) * height * IcecastPlayer.VERTICAL_SCALE)) / 2;
        // Start at exact pixel position to avoid aliasing
        this.ctx.moveTo(0, alignToPixel(firstPoint));
        
        // Draw smooth curve for the active frequencies
        for (let i = 1; i < frequencyRange; i++) {
            const dataIndex = this.activeFreqStart + i;
            const y = (height - ((data[dataIndex] / 255) * height * IcecastPlayer.VERTICAL_SCALE)) / 2;
            
            // Use bezier curve for even smoother line with less aliasing
            if (i > 1 && i < frequencyRange - 1) {
                const prevX = x - sliceWidth;
                const nextX = x + sliceWidth;
                const prevY = (height - ((data[dataIndex - 1] / 255) * height * IcecastPlayer.VERTICAL_SCALE)) / 2;
                const nextY = (height - ((data[dataIndex + 1] / 255) * height * IcecastPlayer.VERTICAL_SCALE)) / 2;
                
                // Control points for bezier curve (1/4 distance to adjacent points)
                const cp1x = prevX + (x - prevX) * 0.75;
                const cp1y = prevY + (y - prevY) * 0.75;
                const cp2x = x + (nextX - x) * 0.25;
                const cp2y = y + (nextY - y) * 0.25;
                
                this.ctx.bezierCurveTo(cp1x, alignToPixel(cp1y), cp2x, alignToPixel(cp2y), x, alignToPixel(y));
            } else {
                // Use quadraticCurveTo for edge points
                const xc = x + sliceWidth / 2;
                this.ctx.quadraticCurveTo(x, alignToPixel(y), xc, alignToPixel(y));
            }
            
            x += sliceWidth;
        }
        
        // Mirror the path to create a symmetric waveform with the same anti-aliasing improvements
        for (let i = frequencyRange - 1; i >= 0; i--) {
            const dataIndex = this.activeFreqStart + i;
            const y = height - (height - ((data[dataIndex] / 255) * height * IcecastPlayer.VERTICAL_SCALE)) / 2;
            
            if (i < frequencyRange - 2 && i > 0) {
                const prevX = x + sliceWidth;
                const nextX = x - sliceWidth;
                const prevY = height - (height - ((data[dataIndex + 1] / 255) * height * IcecastPlayer.VERTICAL_SCALE)) / 2;
                const nextY = height - (height - ((data[dataIndex - 1] / 255) * height * IcecastPlayer.VERTICAL_SCALE)) / 2;
                
                // Control points for bezier curve
                const cp1x = prevX + (x - prevX) * 0.75;
                const cp1y = prevY + (y - prevY) * 0.75;
                const cp2x = x + (nextX - x) * 0.25;
                const cp2y = y + (nextY - y) * 0.25;
                
                this.ctx.bezierCurveTo(cp1x, alignToPixel(cp1y), cp2x, alignToPixel(cp2y), x, alignToPixel(y));
            } else {
                // Use quadraticCurveTo for edge points
                const xc = x - sliceWidth / 2;
                this.ctx.quadraticCurveTo(x, alignToPixel(y), xc, alignToPixel(y));
            }
            
            x -= sliceWidth;
        }
        
        this.ctx.closePath();
    }
    
    /**
     * Handle buffering state
     */
    startBuffering() {
        this.isBuffering = true;
        
        if (this.animationId) {
            cancelAnimationFrame(this.animationId);
            this.animationId = null;
        }
        
        // Update UI to show loading animation
        this.updatePlaybackUI();
    }
    
    stopBuffering() {
        this.isBuffering = false;
        
        if (this.animationId) {
            cancelAnimationFrame(this.animationId);
            this.animationId = null;
        }
        
        // Update UI based on current state
        this.updatePlaybackUI();
    }
    
    /**
     * Play audio function
     */
    playAudio() {
        // Try to preload audio by setting preload attribute
        this.audioElement.preload = "auto";
        
        // Only start buffering animation if we don't have audio data ready
        if (this.audioElement.readyState < 3) {
            this.startBuffering();
        }
        
        // Volume fix for all browsers
        if (this.audioElement.volume !== 1) {
            this.audioElement.volume = 1;
        }
        
        const canUseWebAudio = !this.analyser && this.setupAudioContext();
        
        // Resume audio context first if it's suspended
        if (this.audioContext && this.audioContext.state === 'suspended') {
            console.log('Resuming suspended AudioContext');
            this.audioContext.resume().then(() => {
                console.log('AudioContext resumed successfully');
                this.startPlayback();
            }).catch(error => {
                console.error('Failed to resume AudioContext:', error);
                this.startPlayback();
            });
        } else {
            this.startPlayback();
        }
    }
    
    startPlayback() {
        // Try to play
        this.audioElement.play().then(() => {
            this.isPlaying = true;
            
            // If this was our first autoplay attempt and it was muted, 
            // try unmuting after a short delay
            if (!this.attemptedAutoplay) {
                this.attemptedAutoplay = true;
                setTimeout(() => {
                    // Try to unmute
                    this.audioElement.muted = false;
                }, 500);
            }
            
            this.stopBuffering();
            this.updatePlaybackUI();
        }).catch(error => {
            console.error('Error playing audio:', error);
            
            // If this was our first autoplay attempt with muted audio and it failed,
            // we're probably dealing with strict Safari policies
            if (!this.attemptedAutoplay) {
                this.attemptedAutoplay = true;
                
                // Draw the visualization in paused state to indicate user interaction is needed
                this.isPlaying = false;
                this.stopBuffering();
            } else {
                // If it's a subsequent attempt, just show the visualization
                this.isPlaying = false;
                this.stopBuffering();
            }
        });
    }
    
    /**
     * Pause audio
     */
    pauseAudio() {
        this.audioElement.pause();
        this.isPlaying = false;
        this.isBuffering = false;
        this.hasBeenManuallyPaused = true;  // Track that user manually paused
        
        if (this.animationId) {
            cancelAnimationFrame(this.animationId);
            this.animationId = null;
        }
        
        // Update UI to show play button
        this.updatePlaybackUI();
    }
    
    /**
     * Attempt to recover the stream
     */
    attemptStreamRecovery() {
        if (!this.needsRecovery()) {
            console.log('Stream appears to be working, no recovery needed');
            return;
        }
        
        if (this.reconnectAttempts >= IcecastPlayer.MAX_RECONNECT_ATTEMPTS) {
            console.warn('Maximum reconnection attempts reached. Giving up.');
            return;
        }
        
        this.lastRecoveryAttempt = Date.now();
        this.recoveryInProgress = true;
        
        this.reconnectAttempts++;
        console.log('Attempting to reconnect to stream (attempt ' + this.reconnectAttempts + '/' + IcecastPlayer.MAX_RECONNECT_ATTEMPTS + ')...');
        
        if (this.reconnectTimeout) {
            clearTimeout(this.reconnectTimeout);
        }
        
        const sourceEl = this.audioElement.querySelector('source');
        if (sourceEl && sourceEl.src) {
            const currentSrc = sourceEl.src;
            
            // Remove all query parameters and add a single timestamp
            const baseUrl = currentSrc.split('?')[0];
            const freshSrc = baseUrl + '?t=' + Date.now();
            
            sourceEl.src = freshSrc;
            this.audioElement.load();
            
            const playPromise = this.audioElement.play();
            if (playPromise !== undefined) {
                playPromise.then(() => {
                    console.log('Stream recovered successfully');
                    this.isPlaying = true;
                    this.reconnectAttempts = 0;
                    this.reconnectDelay = IcecastPlayer.INITIAL_RECONNECT_DELAY; // Reset delay
                    this.recoveryInProgress = false;
                    
                    // Update UI to show playing state
                    this.updatePlaybackUI();
                }).catch(error => {
                    console.error('Failed to recover stream:', error);
                    this.reconnectDelay = Math.min(this.reconnectDelay * 1.5, 10000);
                    this.recoveryInProgress = false;
                    this.reconnectTimeout = setTimeout(() => this.attemptStreamRecovery(), this.reconnectDelay);
                });
            } else {
                setTimeout(() => {
                    this.recoveryInProgress = false;
                }, 2000);
            }
        } else {
            console.error('No valid source element found for audio stream');
            this.recoveryInProgress = false;
        }
    }
    
    /**
     * Check if the stream needs recovery
     */
    needsRecovery() {
        const now = Date.now();
        
        // Don't try to recover too frequently
        if (now - this.lastRecoveryAttempt < IcecastPlayer.RECOVERY_DEBOUNCE) {
            return false;
        }
        
        // Don't try to recover if another recovery is already in progress
        if (this.recoveryInProgress) {
            return false;
        }
        
        // Check if audio is currently producing sound
        if (this.audioElement.currentTime > 0 && 
            !this.audioElement.paused && 
            !this.audioElement.ended && 
            this.audioElement.readyState > 2) {
            
            // Check if the currentTime is actually advancing (no stalls)
            if (this.audioElement.currentTime !== this.lastPlaybackTime) {
                // Stream seems to be working correctly, don't try to recover
                return false;
            }
        }
        
        // Stream is in a problematic state
        return true;
    }
}
